<?php

namespace App\Console\Commands;

use Carbon\Carbon;
use App\Models\Invoice;
use Illuminate\Console\Command;
use App\Actions\Tec\PrepareOrder;
use Spatie\Activitylog\Contracts\Activity;

class CreateRecurringInvoices extends Command
{
    protected $description = 'Create recurring invoices';
    protected $signature = 'generate:invoices';

    public function __construct()
    {
        parent::__construct();
    }

    public function handle()
    {
        $recurring_invoices = Invoice::withoutGlobalScopes()
            ->with(['items', 'items.taxes', 'customer:id,name'])->recurring()
            ->where(function ($query) {
                $query->whereNull('next_create_date')->orWhere('next_create_date', '<=', now()->format('Y-m-d'));
            })->cursor()->filter(function ($recurring_invoice) {
                $date = $recurring_invoice->last_created_at ? $recurring_invoice->last_created_at : $recurring_invoice->date;
                if ($recurring_invoice->create_before) {
                    $date = 1 == $recurring_invoice->create_before ? $date->subDay() : $date->subDays($recurring_invoice->create_before);
                }
                return $this->{$recurring_invoice->repeat}($date);
            });

        $recurring_invoices->each(function ($recurring_invoice) {
            $this->create($recurring_invoice);
            $recurring_invoice->disableLogging();
            $recurring_invoice->update([
                'last_created_at'  => date('Y-m-d'),
                'next_create_date' => $this->nextDate($recurring_invoice),
            ]);

            activity()->withProperties($recurring_invoice)
                ->tap(function (Activity $activity) use ($recurring_invoice) {
                    $activity->account_id = $recurring_invoice->account_id;
                })
                ->log(__choice('New invoice has been created for recurring of invoice :invoice for customer :customer and next invoice will be on :date', [
                    'invoice'  => __('number') . ' ' . $recurring_invoice->id . ' (' . __('Ref') . ': ' . $recurring_invoice->reference . ')',
                    'customer' => $recurring_invoice->customer->name . ' (' . __('Id') . ': ' . $recurring_invoice->customer->id . ')',
                    'date'     => $recurring_invoice->next_create_date,
                ]));
        });

        if ($recurring_invoices->count()) {
            $this->info(sprintf('%d ' . __choice('invoice', [], $recurring_invoices->count()) . ' have been created with generate:invoices command.', $recurring_invoices->count()));
        } else {
            $this->info('Nothing to create with generate:invoices command.');
        }
    }

    protected function annually($date)
    {
        return $date->startOfDay()->addYear()->lte(Carbon::now()->startOfDay());
    }

    protected function biennially($date)
    {
        return $date->startOfDay()->addYears(2)->lte(Carbon::now()->startOfDay());
    }

    protected function create($recurring_invoice)
    {
        $v = $recurring_invoice->toArray();
        // $v = unserialize(serialize($recurring_invoice->toArray()));

        $invoice_items = [];

        foreach ($recurring_invoice->items as $item) {
            $invoice_item = [
                'name'                  => $item->name,
                'price'                 => $item->price,
                'quantity'              => $item->quantity,
                'net_price'             => $item->net_price,
                'unit_price'            => $item->unit_price,
                'account_id'            => $item->account_id,
                'product_id'            => $item->product_id ?? null,
                'details'               => $item->details,
                'discount_amount'       => $item->discount_amount,
                'total_discount_amount' => $item->total_discount_amount,
                'discount'              => $item->discount,
                'tax_method'            => $item->tax_method,
                'tax_amount'            => $item->tax_amount,
                'total_tax_amount'      => $item->total_tax_amount,
                'extra_attributes'      => $item->extra_attributes,
                'total'                 => $item->total,
            ];

            $invoice_item['taxes'] = [];
            if (!empty($item->taxes)) {
                // foreach ($item->taxes as $d) {
                //     $invoice_item['taxes'][] = $d->id;
                // }
                $invoice_item['taxes'][] = $item->taxes->pluck('id')->all();
            }

            // dump($invoice_item);
            $invoice_items[] = $invoice_item;
        }

        $v['items'] = $invoice_items;
        $v['date'] = Carbon::now()->format('Y-m-d');
        $v['invoice_id'] = $recurring_invoice->id;
        $v['account_id'] = $recurring_invoice->account_id;
        unset(
            $v['id'],
            $v['repeat'],
            $v['recurring'],
            $v['create_before'],
            $v['last_created_at'],
            $v['hash'],
            $v['reference'],
            $v['created_at'],
            $v['updated_at'],
            $v['attachment']
        );

        $invoice = (new PrepareOrder($v, new Invoice()))->save();
        // charge payment method
    }

    protected function daily($date)
    {
        return $date->startOfDay()->addDay()->lte(Carbon::now()->startOfDay());
    }

    protected function monthly($date)
    {
        return $date->startOfDay()->addMonth()->lte(Carbon::now()->startOfDay());
    }

    protected function nextDate($invoice)
    {
        return match ($invoice->repeat) {
            'daily'      => $invoice->last_created_at->addDay()->format('Y-m-d'),
            'weekly'     => $invoice->last_created_at->addWeek()->format('Y-m-d'),
            'monthly'    => $invoice->last_created_at->addMonth()->format('Y-m-d'),
            'quarterly'  => $invoice->last_created_at->addQuarter()->format('Y-m-d'),
            'semiannual' => $invoice->last_created_at->addMonths(6)->format('Y-m-d'),
            'annually'   => $invoice->last_created_at->addYear()->format('Y-m-d'),
            'biennially' => $invoice->last_created_at->addYears(2)->format('Y-m-d'),
        };
    }

    protected function quarterly($date)
    {
        return $date->startOfDay()->addQuarter()->lte(Carbon::now()->startOfDay());
    }

    protected function semiannual($date)
    {
        return $date->startOfDay()->addMonths(6)->lte(Carbon::now()->startOfDay());
    }

    protected function weekly($date)
    {
        return $date->startOfDay()->addWeek()->lte(Carbon::now()->startOfDay());
    }
}
