<?php

namespace App\Http\Controllers;

use Inertia\Inertia;
use Illuminate\Http\Request;
use App\Actions\Tec\PrepareOrder;
use App\Http\Requests\InvoiceRequest;
use App\Http\Resources\InvoiceCollection;
use App\Models\{Company, Field, Invoice, Note, Quotation, TaxRate};

class InvoiceController extends Controller
{
    public function create(Request $request)
    {
        $quote = null;
        if ($request->from_quote) {
            $quote = Quotation::with(['company:id,name', 'customer:id,name', 'items', 'taxes:id,name', 'items.taxes:id,name'])->findOrFail($request->get('from_quote'));
        }
        return Inertia::render('Invoice/Form', [
            'quote'           => $quote,
            'customer_fields' => Field::ofModel('customer')->get(),
            'item_fields'     => Field::ofModel('invoiceitem')->get(),
            'fields'          => Field::ofModel('maininvoice')->get(),
            'note'            => Note::where('default_sale', 1)->first(),
            'companies'       => Company::get(['id as value', 'name as label']),
            'tax_rates'       => TaxRate::all(['id as value', 'name as label', 'id', 'rate', 'fixed']),
        ]);
    }

    public function destroy(Request $request, Invoice $invoice)
    {
        if ($invoice->{$request->force ? 'forceDelete' : 'delete'}()) {
            return redirect()->route('invoices.index')->with('message', __choice('{record} has been {action}.', ['record' => __('Invoice'), 'action' => __('deleted')]));
        }
        return back()->with('error', __('The record can not be deleted.'));
    }

    public function destroyMany(Request $request)
    {
        $count = $failed = 0;
        foreach (Invoice::whereIn('id', $request->selection)->get() as $invoice) {
            if ($invoice->{$request->force ? 'forceDelete' : 'delete'}()) {
                $count++;
            } else {
                $failed++;
            }
        }
        return back()->with('message', __('The task has completed, {count} deleted and {failed} failed.', ['count' => $count, 'failed' => $failed]));
    }

    public function destroyPermanently(Invoice $invoice)
    {
        if ($invoice->forceDelete()) {
            return redirect()->route('invoices.index')->with('message', __choice('{record} has been {action}.', ['record' => __('Invoice'), 'action' => __('permanently deleted')]));
        }
        return back()->with('error', __('The record can not be deleted.'));
    }

    public function edit(Invoice $invoice)
    {
        return Inertia::render('Invoice/Form', [
            'customer_fields' => Field::ofModel('customer')->get(),
            'item_fields'     => Field::ofModel('invoiceitem')->get(),
            'fields'          => Field::ofModel('maininvoice')->get(),
            'note'            => Note::where('default_sale', 1)->first(),
            'companies'       => Company::get(['id as value', 'name as label']),
            'tax_rates'       => TaxRate::all(['id as value', 'name as label', 'id', 'rate', 'fixed']),
            'edit'            => ['data' => $invoice->load(['company', 'customer', 'items', 'user', 'taxes', 'items.taxes'])],
        ]);
    }

    public function index(Request $request)
    {
        $filters = $request->all('search', 'recurring', 'start_date', 'end_date', 'trashed');

        return Inertia::render('Invoice/List', [
            'filters'        => $filters,
            'payment_fields' => Field::ofModel('payment')->get(),
            'invoices'       => new InvoiceCollection(
                Invoice::with(['company', 'customer', 'items', 'user'])->filter($filters)->orderByDesc('date')->latest()->paginate()->onEachSide(2)->withQueryString()
            ),
        ]);
    }

    public function restore(Invoice $invoice)
    {
        $invoice->restore();
        $invoice->items->each->restore();
        return back()->with('message', __choice('{record} has been {action}.', ['record' => __('Invoice'), 'action' => __('restored')]));
    }

    public function show(Invoice $invoice)
    {
        return Inertia::render('Invoice/Details', [
            'invoice' => $invoice->load(['company', 'customer', 'items', 'user', 'taxes', 'items.taxes']),
        ]);
    }

    public function store(InvoiceRequest $request)
    {
        $invoice = (new PrepareOrder($request->validated(), new Invoice()))->save();
        $invoice->saveAttachments($request->file('attachments'));
        return redirect()->route('invoices.index')
            ->with('message', __choice('{record} has been {action}.', ['record' => __('Invoice'), 'action' => __('created')]));
    }

    public function update(InvoiceRequest $request, Invoice $invoice)
    {
        $invoice = (new PrepareOrder($request->validated(), $invoice, true))->save();
        $invoice->saveAttachments($request->file('attachments'));
        return redirect()->route('invoices.index')
            ->with('message', __choice('{record} has been {action}.', ['record' => __('Invoice'), 'action' => __('updated')]));
    }
}
