<?php

namespace App\Http\Controllers;

use Inertia\Inertia;
use App\Models\Setting;
use App\Rules\LocaleLength;
use Illuminate\Http\Request;
use Illuminate\Support\Facades\File;
use Illuminate\Support\Facades\Storage;

class SettingController extends Controller
{
    public function index()
    {
        return Inertia::render('Setting/Index', ['current' => get_settings()]);
    }

    public function store(Request $request)
    {
        $langFiles = collect(json_decode(File::get(lang_path('languages.json')))->available)->pluck('value')->all();
        $settings = $request->validate([
            'name'             => 'required',
            'currency_code'    => 'required|string|size:3',
            'language'         => 'required|string|in:' . implode(',', $langFiles),
            'per_page'         => 'required|in:10,15,25,50,100',
            'default_locale'   => ['required', new LocaleLength()],
            'fraction'         => 'required|numeric|min:0|in:0,1,2,3,4',
            'invoice_status'   => 'required|string|in:badge,text',
            'initial_rows'     => 'required|numeric|min:1|max:20',
            'reference'        => 'required|string|in:ai,ulid,uuid,uniqid',
            'paypal'           => 'nullable|array',
            'stripe'           => 'nullable|array',
            'hour12'           => 'nullable|boolean',
            'paypal.*'         => 'nullable',
            'stripe.*'         => 'nullable',
            'product_discount' => 'nullable|boolean',
            'create_user'      => 'nullable|boolean',
            'over_selling'     => 'nullable|boolean',
            'impersonation'    => 'nullable|boolean',
            'login_logo'       => 'nullable|mimes:png,jpg,jpeg,svg',
            'invoice_logo'     => 'nullable|mimes:png,jpg,jpeg,svg',
            'attachment_exts'  => 'nullable',
        ]);

        if (demo()) {
            $settings['name'] = 'SIM';
            $settings['login_logo'] = asset('/storage/logo.svg');
            $settings['invoice_logo'] = asset('/storage/logo.svg');
        } else {
            if ($request->has('login_logo') && $request->login_logo) {
                $settings['login_logo'] = Storage::disk('assets')->url($request->login_logo->store('/images', 'assets'));
            }
            if ($request->has('invoice_logo') && $request->invoice_logo) {
                $settings['invoice_logo'] = Storage::disk('assets')->url($request->invoice_logo->store('/images', 'assets'));
            }
        }

        $json_fields = ['paypal', 'stripe'];
        collect($settings)->each(function ($value, $key) use ($json_fields) {
            Setting::updateOrCreate(['tec_key' => $key, 'account_id' => auth()->user()->account_id], ['tec_value' => in_array($key, $json_fields) ? json_encode($value) : $value]);
        });
        log_activity(__choice('{record} has been {action}.', ['record' => 'Settings', 'action' => 'saved']), $settings, auth()->user(), 'Setting');

        return back()->with('message', __choice('{record} has been {action}.', ['record' => 'Settings', 'action' => 'saved']));
    }
}
